#!/usr/bin/env python
from __future__ import print_function

import os
import apt.cache

from apt_pkg import version_compare
from apt.progress.text import OpProgress

class PkgCache(object):

    def __init__(self, pkgnames=set(), skip_pkgnames=set(), autoupdate=True,
            show_installed=False, show_upgradable=True, debug=False):

        self.cache = apt.cache.Cache()
        if (autoupdate == True): 
            self.update()

        """ config parameters """
        self.show_installed = show_installed
        self.show_upgradable = show_upgradable
        self.debug = debug

        """ initialize data structures """
        self.provides  = set()
        self.requires  = set()
        self.installed = set()
        self.upgrades  = set()
        self.skippkgs  = set()

        """ get packages data """
        if len(pkgnames) > 0:
            self.__get_provides(pkgnames)
            self.__get_depends(pkgnames)
            self.__get_installed(pkgnames)
            self.__get_upgrades(pkgnames)
            self.__get_requires(pkgnames)
        if len(skip_pkgnames) > 0:
            self.__get_skippkgs(skip_pkgnames)


    """ get a set of packages that each package depends """
    def __get_depends(self, pkgnames=set()):
        if len(self.provides) <= 0:
            self.__get_provides(pkgnames)

        for package in self.provides:
            candver = self.cache._depcache.get_candidate_ver(package._pkg)
            if candver is None:
                continue
            dependslist = candver.depends_list
            if 'Depends' in dependslist:
                for depVerList in dependslist['Depends']:
                    for dep in depVerList:
                        if (dep.target_pkg.name in self.cache and
                            dep.target_pkg.name != package.name):
                            self.requires.add(self.cache[dep.target_pkg.name])

    """ get a set of installed packages """
    def __get_installed(self, pkgnames=set()):
        if len(self.requires) <= 0:
            self.__get_depends(pkgnames)

        for package in self.requires:
            if package.is_installed:
                self.installed.add(package)

    """ get a set of what each (meta/virtual) package provides """
    def __get_provides(self, pkgnames=set()):
        self.cache.open(OpProgress())
        for pkgname in pkgnames:
            if not self.cache.is_virtual_package(pkgname):
                self.provides.add(self.cache[pkgname])
            else: # it is a virtual package
                provide = None
                for package in self.cache.get_providing_packages(pkgname):
                    if provide == None:
                        provide = package
                    elif version_compare(package.versions._versions[0].ver_str,
                                         provide.versions._versions[0].ver_str) > 0:
                        provide = package
                    else:
                        continue
                self.provides.add(provide)
        self.cache.close()

    """ add packages to requires set """
    def __get_requires(self, pkgnames=set()):
        for pkgname in pkgnames:
            if not self.cache.is_virtual_package(pkgname):
                self.requires.add(self.cache[pkgname])


    """ get a set of packages that will be skipped """
    def __get_skippkgs(self, pkgnames=set()):
        for pkgname in pkgnames:
            if not self.cache.is_virtual_package(pkgname):
                self.skippkgs.add(self.cache[pkgname])

    """ get a set of upgradable packages """
    def __get_upgrades(self, pkgnames=set()):
        if len(self.installed) <= 0:
            self.__get_installed(pkgnames)

        for package in self.installed:
            if package.is_upgradable:
                self.upgrades.add(package)

    @staticmethod
    def __print_summary(data):
        r, c = os.popen('stty size', 'r').read().split()
        nc = int(c) - 5
        cs = (nc - 2) / 4

        hdr = ['Act.  Package Name','Installed','Available','Architecture']
        line_fmt = "{: <%d} " * 5 % tuple([2]+[cs for i in range(0,len(hdr))])

        def __println(str='-'):
            print ("%s" % (str * int(nc)))

        def __header():
            print ("\nThe following package changes will be performed:")
            __println()
            print (line_fmt.format(*(['']+hdr)), end='')
            __println()

        def __print():
            __header()
            for row in data:
                print (line_fmt.format(*(['']+row)), end='')
            __println(' ')
        __print()


    """ get information of what operation will be done """
    def __summarize(self):
        data = []
        pkgset = self.requires

        if self.show_installed == False:
            """ show installable/upgradable packages """
            pkgset = pkgset - (self.installed - self.upgrades)

        if self.show_upgradable == False:
            """ show installable packages only """
            pkgset = pkgset - self.upgrades

        """ remove skipped packages from package set """
        self.pkgset = pkgset - self.skippkgs

        for package in sorted(self.pkgset):
            cur_version = '-'
            action = 'in'
            if package.is_installed:
                cur_version = package.installed.version
                action = '--'
                if package.is_upgradable:
                    try:
                        search_f = lambda f: f.name == package.name
                        result_f = filter(search_f, self.installed)
                        inst_pkg = result_f[0].installed
                        cur_version = inst_pkg.version
                    except IndexError:
                        cur_version = package.installed.version
                    action = 'up'

            data.append([
                '[%s]  %s' % (action, package.candidate._cand.parent_pkg.name),
                cur_version,
                package.candidate._cand.ver_str,
                package.candidate._cand.arch
            ])
        self.__print_summary(data)

    def __get_response(self):
        if self.debug:
            return False
        return True

    def __mark_install(self):
        for pkg in self.pkgset:
            if self.debug:
                print("-> Marking pkg for install: %s" % pkg.name)
            pkg.mark_install()

    def install(self):
        self.__summarize()
        if self.__get_response() == True:
            self.cache.open(OpProgress())
            self.__mark_install()
            self.cache.commit()
            self.cache.close()


    def update(self):
        from apt.progress.text import AcquireProgress
        self.cache.update(AcquireProgress())

# vim:number:sw=4:sts=4:ts=4:et:autoindent:syntax=python
